#include "GraphicView.h"
#include "constants.h"

#include <Directory.h>
#include <TranslationUtils.h>
#include <Window.h>
#include <Bitmap.h>
#include <stdio.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <algorithm>
#include <Path.h>
#include <String.h>
#include <Alert.h>

GraphicView::GraphicView(BRect frame, const char *name, entry_ref initialDirectory)
	: BView(frame, name, B_FOLLOW_ALL_SIDES, B_WILL_DRAW | B_FRAME_EVENTS | B_PULSE_NEEDED)
{
	fBitmap = NULL;
	fMenu = NULL;
	fMode = ORIGINAL_SIZE;
	fKeyDown = false;
	fRef = initialDirectory;
	fCollector = new FileCollector(fRef, "image");
	CreateMenu();
	RestoreState();
	Reset(fRef);
}

GraphicView::~GraphicView()
{
	SaveState();
	delete fCollector;
	delete fMenu;
}

void GraphicView::AttachedToWindow()
{
	MakeFocus();
	//現在のディレクトリを保存
	entry_ref currentRef = fCollector->GetDirectory();
	//画像を探す
	status_t sts = NextImage();
	if (sts != B_OK)
	{
		//なかったら子フォルダを探す
		sts = ChildDirectory();
		//子フォルダがあったら
		if (sts == B_OK)
			//画像を探す
			sts = NextImage();
		//子フォルダがない、または子フォルダに画像がない場合
		if (sts != B_OK)
		{
			//あきらめて最初のフォルダを表示
			Reset(currentRef);
			NextImage();
		}
	}
		
}

void GraphicView::Draw(BRect updateRect)
{
	BView::Draw(updateRect);
	if (fBitmap != NULL)
	{
		BRect bitmapRect = fBitmap->Bounds();
		BRect viewRect = this->Bounds();
		BRect destination = bitmapRect;
		switch (fMode)
		{
			case ORIGINAL_SIZE:
			{
				break;
			}
			case SCALE:
			{
				float bitmapRatio = (bitmapRect.Height() + 1)/(bitmapRect.Width() + 1);
				float viewRatio = (viewRect.Height() + 1)/(viewRect.Width() + 1);
				float expandRatio;
				if (bitmapRatio > viewRatio)
				{
					expandRatio = (viewRect.Height() + 1)/(bitmapRect.Height() + 1);
				}
				else
				{
					expandRatio = (viewRect.Width() + 1)/(bitmapRect.Width() + 1);
				}
				destination.right = destination.left + bitmapRect.Width()*expandRatio;
				destination.bottom = destination.top + bitmapRect.Height()*expandRatio;
				break;
			}
		}
		destination.OffsetTo((viewRect.Width() - destination.Width())/2, (viewRect.Height() - destination.Height())/2);
		DrawBitmap(fBitmap, destination);
	}
}

void GraphicView::Pulse()
{
	//キーを押しっぱなしにしているときに、時々画像を移動させてやる
	fKeyDown = false;
	
}

void GraphicView::KeyUp(const char *bytes, int32 numBytes)
{
	BView::KeyUp(bytes, numBytes);
	fKeyDown = false;
}

//
void GraphicView::KeyDown(const char *bytes, int32 numBytes)
{
	BView::KeyDown(bytes, numBytes);
	if (numBytes == 1)
	{
		if (fKeyDown == true) return;
		fKeyDown = true;
		uint32 mod = modifiers();
		//もとはB_COMMAND_KEYにしていたんだけど
		//どうやらB_COMMAND_KEYを押しているときはKeyDownが呼び出されないようだ。
		//というわけでB_CONTROL_KEYに変えた。
		bool commandOn = mod & B_CONTROL_KEY;
		switch (bytes[0])
		{
			//次に移動
			case B_SPACE:
			case B_RIGHT_ARROW:
			case B_ENTER:
			case B_PAGE_DOWN:
			{
				//コマンドキーを押している場合は最後に移動
				if (commandOn)
				{
					LastImage();
				}
				else
				{
					NextImage();
				}
				break;
			}
			//前に戻る
			case B_BACKSPACE :
			case B_LEFT_ARROW :
			case B_PAGE_UP :
			{
				//コマンドキーを押している場合は先頭に移動
				if (commandOn)
				{
					FirstImage();
				}
				else
				{
					PreviousImage();
				}
				break;
			}
			//先頭に移動
			case B_HOME:
			{
				FirstImage();
				break;
			}
			//最後に移動
			case B_END:
			{
				LastImage();
				break;
			}
			//前のフォルダ
			case B_UP_ARROW :
			{	
				if (commandOn)
				{
					//親フォルダ
					ShowParentDirectory();
				}
				else
				{
					//前のフォルダに移動したときはフォルダの最後を表示
					ShowPreviousDirectory();
				}
				break;
			}
			//次のフォルダ
			case B_DOWN_ARROW :
			{	
				if (commandOn)
				{
					ShowChildDirectory();
				}
				else
				{
					ShowNextDirectory();
				}
				break;
			}
			//終了
			case 'q' :
			case 'w' :
				Window()->PostMessage(B_QUIT_REQUESTED);
				break;
			case 'f':
				//dynamic_cast<AlesWindow*>(Window())->FullScreen();//PostMessage(FULL_SCREEN);
				Window()->PostMessage(FULL_SCREEN);
				break;
			default :
			{
				fKeyDown = false;//関係のないキーを押した場合はキーダウンをなしにする
				break;
			}
		}
	}
}

void GraphicView::MessageReceived(BMessage *message)
{
	if (message == NULL) return;
	switch (message->what)
	{
		//マウスぐりぐり
		case B_MOUSE_WHEEL_CHANGED:
		{
			MouseWheelChanged(message);
			break;
		}
		//画像の並べ替え
		case NAME :
		case SIZE :
		case CREATION_TIME :
		case MODIFICATION_TIME :
		case ACCESS_TIME :
		{
			fCollector->SortBy(message->what);
			NextImage();
			break;
		}
		//画像表示方法
		case ORIGINAL_SIZE:
		case SCALE:
		{
			fMode = message->what;
			Invalidate();
			break;
		}
		case NEXT_IMAGE:
		{
			NextImage();
			break;
		}
		case PREVIOUS_IMAGE:
		{
			PreviousImage();
			break;
		}
		case LAST_IMAGE:
		{
			LastImage();
			break;
		}
		case NEXT_DIRECTORY:
		{
			ShowNextDirectory();
			break;
		}
		case PREVIOUS_DIRECTORY:
		{
			ShowPreviousDirectory();
			break;
		}
		case PARENT_DIRECTORY:
		{
			ShowParentDirectory();
			break;
		}
		case CHILD_DIRECTORY:
		{
			ShowChildDirectory();
			break;
		}
		//ヘルプダイアログ
		case HELP:
		{
			const char text[] =
				"Ales by muramoto\n"
				"Inspired from Sela\n"
				"\n"
				"->, SPACE, ENTER : Next Image\n"
				"<-, BACKSPACE : Previous Image\n"
				"Alt(Ctl) + -> : Last Image\n"
				"Alt(Ctl) + <- : First Image\n"
				"DOWN ARROW : Next Folder\n"
				"UP ARROW : Previous Folder\n"
				"Alt(Ctl) + DOWN ARROW : Child Folder\n"
				"Alt(Ctl) + UP ARROW : Parent Folder\n"
				"Q, W : Quit\n"
				"\n"
				"E-mail : t-muramoto@geocities.co.jp";
			(new BAlert("Ales", text, "OK"))->Go();
			break;
		}
		case B_QUIT_REQUESTED :
			Window()->PostMessage(B_QUIT_REQUESTED);
			break;
		default :
		{
			BView::MessageReceived(message);
			break;
		}
	}
}

//ウィンドウをリサイズしたときは画像も追従する
void GraphicView::FrameResized(float width, float height)
{
	Invalidate();
}

//ポップアップメニューを表示
void GraphicView::MouseDown(BPoint point)
{
	BMenuItem *selected;
	
	ConvertToScreen(&point);
	point.x--; point.y--;//カーソルとメニューが重なるようにする
	selected = fMenu->Go(point, false, true, BRect(0,0,10,10));
	
	if (selected)
	{
		BMessage *message = selected->Message();
		//マークされているアイテムがあったらラジオモードということにして
		//マークを外して、選択したアイテムにマークをつける
		//本当はBMenu::SetRadioMode(true)にしたいんだけど、
		//どうも有効にならないようだ。
		BMenu *selectedMenu = selected->Menu();
		BMenuItem *markedItem = selectedMenu->FindMarked();
		if (markedItem != NULL)
		{
			markedItem->SetMarked(false);
			selected->SetMarked(true);
		}
		MessageReceived(message);
	}
}

//フォルダ内の先頭の画像
status_t GraphicView::FirstImage()
{
	fCollector->ResetPosition();
	return NextImage();
}

//フォルダ内の最後の画像
status_t GraphicView::LastImage()
{
	fCollector->ResetPosition();
	return PreviousImage();
}

//次のフォルダ
status_t GraphicView::ShowNextDirectory()
{
	status_t sts = ChangeDirectory(1);
	if (sts == B_OK)
		return NextImage();
	return sts;
}

//前のフォルダ
status_t GraphicView::ShowPreviousDirectory()
{
	status_t sts = ChangeDirectory(-1);
	if (sts == B_OK)
		return PreviousImage();
	return sts;
}

//親フォルダ
status_t GraphicView::ShowParentDirectory()
{
	status_t sts = ParentDirectory();
	if (sts == B_OK)
		return NextImage();
	return sts;
}

//子フォルダ
status_t GraphicView::ShowChildDirectory()
{
	status_t sts = ChildDirectory();
	if (sts == B_OK)
		return NextImage();
	return sts;
}

//次の画像に移る
status_t GraphicView::NextImage()
{
	entry_ref ref;
	//次のファイルを取得
	status_t sts = fCollector->GetNextRef(&ref);
	if (sts != B_OK)
	{
		return sts;
		/*
		//なかったら下のフォルダに移動
		if (ChildDirectory() != B_OK)
		{
			//出来なかったら隣のフォルダに移動
			if (NextDirectory() != B_OK)
			{
				//できなかったらどうしよう
				return;
			}
		}
		*/
	}
	ShowImage(ref);
	return B_OK;
}

//前の画像に移る
status_t GraphicView::PreviousImage()
{
	entry_ref ref;
	//前のファイルを取得
	status_t sts = fCollector->GetPrevRef(&ref);
	if (sts != B_OK)
	{
		return sts;
		/*
		//なかったら前のフォルダに移動
		if (PreviousDirectory() != B_OK)
		{
			//出来なかったら隣のフォルダに移動
			if (ParentDirectory() != B_OK)
			{
				//できなかったらどうしよう
				return;
			}
		}
		*/
	}
	ShowImage(ref);
	return B_OK;
}

//同じ階層でディレクトリを移動
status_t GraphicView::ChangeDirectory(int32 moveBy)
{
	status_t sts;
	entry_ref dir = fCollector->GetDirectory();
	BEntry me(&dir), parent;
	sts = me.GetParent(&parent);
	if (sts != B_OK) return sts;
	entry_ref parentRef;
	sts = parent.GetRef(&parentRef);
	if (sts != B_OK) return sts;
	//自分を探す
	FileCollector dirCollector(parentRef, "application/x-vnd.Be-directory");
	dirCollector.SortBy(fMenu->FindItem("Sort By")->Submenu()->FindMarked()->Message()->what);
	dirCollector.SetPosition(dir);
	entry_ref ref;
	//次に進む
	//今のところ一つだけしか移動できない
	if (moveBy >=0)
	{
		//次のディレクトリを取得するはず
		sts = dirCollector.GetNextRef(&ref);
		if (sts != B_OK) return sts;
	}
	//前に戻る
	else
	{
		//一つ前のを取得するはず
		sts = dirCollector.GetPrevRef(&ref);
		if (sts != B_OK) return sts;
	}
	Reset(ref);
	return B_OK;
}

//子フォルダに移動
status_t GraphicView::ChildDirectory()
{
	status_t sts;
	FileCollector dirCollector(fCollector->GetDirectory(), "application/x-vnd.Be-directory");
	entry_ref ref;
	sts = dirCollector.GetNextRef(&ref);
	if (sts != B_OK) return sts;
	Reset(ref);
	return B_OK;
}

//親フォルダに移動
status_t GraphicView::ParentDirectory()
{
	status_t sts;
	entry_ref currentDir = fCollector->GetDirectory();
	BEntry current(&currentDir), parent;
	sts = current.GetParent(&parent);
	if (sts != B_OK) return sts;
	entry_ref ref;
	sts = parent.GetRef(&ref);
	if (sts != B_OK) return sts;
	Reset(ref);
	return B_OK;
}


//画像を取得する
void GraphicView::ShowImage(entry_ref ref)
{
	/*
	if (fList.empty() == true)
	{
		PrintNoImage();
		return;
	}
	*/
	delete fBitmap;
	fBitmap = BTranslationUtils::GetBitmap(&ref);
	BEntry entry(&ref);
	BPath path;
	if (entry.GetPath(&path) == B_OK)
		Window()->SetTitle(path.Path());
	Invalidate();
}

//フォルダに画像ファイルがない場合にその旨を表示する
void GraphicView::PrintNoImage()
{
	/*
	*/
}

//フォルダを移動したときに画像リストを更新する
void GraphicView::Reset(entry_ref dirRef)
{
	fCollector->ResetTo(dirRef, "image");
	BPath path;
	BEntry entry(&dirRef);
	if (entry.GetPath(&path) == B_OK)
	{
		BString string = path.Path();
		if (fCollector->Count() <= 0)
			string += " -- No Image";
		Window()->SetTitle(string.String());
	}

}
//ポップアップメニューの作成
void GraphicView::CreateMenu()
{
	fMenu = new BPopUpMenu(kPopUpMenu, false, false);
	//画像配置メニュー
	BMenu *mode = new BMenu(kMode);
	mode->AddItem(new BMenuItem(kOriginalSize, new BMessage(ORIGINAL_SIZE)));
	mode->AddItem(new BMenuItem(kScale, new BMessage(SCALE)));
	fMenu->AddItem(mode);
	//ソートメニュー
	BMenu *sort = new BMenu(kSortBy);
	sort->SetRadioMode(true);
	sort->AddItem(new BMenuItem(kName, new BMessage(NAME)));
	sort->AddItem(new BMenuItem(kSize, new BMessage(SIZE)));
	sort->AddItem(new BMenuItem(kCreationTime, new BMessage(CREATION_TIME)));
	sort->AddItem(new BMenuItem(kModificationTime, new BMessage(MODIFICATION_TIME)));
	sort->AddItem(new BMenuItem(kAccessTime, new BMessage(ACCESS_TIME)));
	fMenu->AddItem(sort);
	//----
	fMenu->AddSeparatorItem();
	//ヘルプ・終了
	fMenu->AddItem(new BMenuItem(kHelp, new BMessage(HELP)));
	fMenu->AddItem(new BMenuItem(kQuit, new BMessage(B_QUIT_REQUESTED)));
}

//設定を読み込み
status_t GraphicView::RestoreState()
{
	SettingsFile setting(kSettingFile, kSettingDir);
	status_t sts = setting.InitCheck();
	//if(sts != B_OK) return sts;
	sts = setting.Load();
	
	//渡されたパスがディレクトリじゃない場合、保存されているパスを読み込む
	BEntry entry(&fRef, true);
	if (entry.IsDirectory() == false)
	{
		BString path;
		entry_ref ref;
		sts = setting.FindString(kPath, &path);
		sts = get_ref_for_path(path.String(), &fRef);
		if(sts != B_OK)
		{
			BPath path;
			find_directory(B_USER_DIRECTORY, &path);
			get_ref_for_path(path.Path(), &fRef);
		}
	}
	
	//モードの読み込み
	RestoreMarked(&setting, kMode, kModeSetting);
	//ソートの読み込み
	RestoreMarked(&setting, kSortBy, kSortSetting);
	return sts;
	

}

//保存してあるモードとかを元に戻す
BMenuItem *GraphicView::RestoreMarked(SettingsFile *setting, const char *label, const char *save_name)
{
	status_t sts;
	int32 saved;
	sts = setting->FindInt32(save_name, &saved);
	//保存してあるのが見つからなかったら0にする
	if (sts != B_OK)
		saved = 0;
	BMenuItem *item;
	item = fMenu->FindItem(label);
	if (item == NULL)
		return NULL;
	int32 index = fMenu->IndexOf(item);
	BMenu *submenu = fMenu->SubmenuAt(index);
	if (submenu == NULL) return NULL;
	BMenuItem *mark = submenu->ItemAt(saved);
	//保存時よりもメニューの数が少なかったりして
	if (mark == NULL)
		mark = submenu->ItemAt(0);
	mark->SetMarked(true);
	//メニューの実行もする
	this->MessageReceived(mark->Message());
	return mark;
}

//マウスぐりぐり対応予定
//持っているマウスがぐりぐりできないので動作確認していない
void GraphicView::MouseWheelChanged(BMessage *message)
{
	float y;
	status_t sts;
	sts = message->FindFloat("be:wheel_delta_y", &y);
	if (sts != B_OK) return;
	if (y > 0)
	{
		NextImage();
	}
	else if (y < 0)
	{
		PreviousImage();
	}
	
}

//設定を保存
status_t GraphicView::SaveState()
{
	SettingsFile setting(kSettingFile, kSettingDir);
	status_t sts = setting.InitCheck();
	if(sts != B_OK)
		return sts;
	sts = setting.Load();
	//現在のパスを保存
	entry_ref ref = fCollector->GetDirectory();
	BEntry entry(&ref);
	BPath path;
	entry.GetPath(&path);
	setting.SetString(kPath, path.Path());
	//ソート順を保存
	SaveMarked(&setting, kSortBy, kSortSetting);
	//画像スケールを保存
	SaveMarked(&setting, kMode, kModeSetting);
	
	sts = setting.Save();
	return sts;
}

void GraphicView::SaveMarked(SettingsFile *setting, const char *label, const char *save_name)
{
	BMenuItem *item;
	item = fMenu->FindItem(label);
	if (item != NULL)
	{
		int32 index = fMenu->IndexOf(item);
		BMenu *submenu = fMenu->SubmenuAt(index);
		if (submenu == NULL) return;
		BMenuItem *marked = submenu->FindMarked();
		if (marked == NULL) return;
		setting->SetInt32(save_name, submenu->IndexOf(marked));
	}

}
